(***********************************************************************)
(*                                                                     *)
(*      bibliotheque flight,   nocf  proto en  Camlp4                  *)
(*                                                                     *)
(*     Baia, Garigou : TER maitrise Paris 6 printemps 2001             *)
(*     modifie par                                                     *)
(*       Emmanuel Chailloux, Equipe PPS, universite P7, projet CARAML  *)
(*                                                                     *)
(*     $Id: flight.ml,v 1.2 2003/04/23 14:25:30 emmanuel Exp $                                                            *)
(*                                                                     *)
(***********************************************************************)


(*----------------------- constantes globales ------------------------*)

(* numero du processus *)
let local = int_of_string Sys.argv.(1);;

(* nombre de processus *)
let nodes = int_of_string Sys.argv.(2);;

(* port du processus local *)
let port = int_of_string Sys.argv.(3);;

(* profondeur d'asynchronisme *)
let wave_depth = int_of_string Sys.argv.(4);;

(* tableau contenant les adresses IP de tous les processus *)
let addrs = 
  Array.init nodes 
    (fun i -> (Unix.gethostbyname (Sys.argv.(i+5))).Unix.h_addr_list.(0));;

(* type des valeurs envoyes *)
type 'a valeur = Valeur of 'a | Exn of exn;;

(* type des fermetures *)
type 'a fermeture = unit -> 'a;;

(* exceptions *)
exception RemoteExn of exn;; (* exception produite  distance *)
exception NestedSync;; (* detection de sync imbriqus *)

(* utilis pour les commentaires *)
let who = "Processus " ^ (string_of_int local) ^ ": ";; 

(* indication de depart par le moniteur *)
let start = ref false;;

(*-------------------- variables et fonctions ------------------------*) 

(* tableau contenant les prochaines vagues des nodes processus*)
let waves = Array.make nodes 1;;
			
(* vague sur laquelle 'surfe' le processus local *)
let wave = ref 0;;
		       
(* plus vielle vague connue *)
let old_wave = ref 1;;
			
(* liste contenant le numero des sync pour les vagues deja commenes *)
let sync_list = ref ([] : int list);;
		       
(* tableau contenant les fermetures *)
let env = ref (Array.make wave_depth ([] : string list));;
let env_buffer = ref ([] : string list);;
			     
(* type de sync *)
let sync_type = [| 0;0 |];;

(* mutex permettant d'acceder au tableau env *)
let e = Mutex.create();;

(* mutex permettant d'acceder au tableau waves *)
let w = Mutex.create();;

(* renvoie le numro de processus  partir d'une sockaddr *)
let processus = function
  | Unix.ADDR_INET(host,_) -> 
      let n = ref 0 in
        for i=0 to nodes-1 do
	  if (addrs.(i) == host) then n := i
	done;
	!n
  | _ -> -1;;
	  
(* ajout de la fermeture au 'buffer' *)
let save_closure c =
    env_buffer := (c::(!env_buffer));;

(* renvoie la plus vieille vague
   = recherche du plus petit element de waves *)
let refresh_old_wave () =
  let min = ref waves.(0) in
    for i = 1 to nodes-1 do
      if waves.(i) < !min then min := waves.(i)
    done;
    old_wave := !min;;

(* indique aux autres processus le fin d'une vague *)
let refresh_waves ()=
  for i=0 to nodes-1 do 
    if (i <> local) 
    then 
      let (ic,oc) = 
        ThreadUnix.open_connection (Unix.ADDR_INET(addrs.(i),port+i-local)) 
      in
        output_string oc ((Marshal.to_string (local, 0, 0) []) ^ "\n");
	flush oc;
        Unix.shutdown_connection ic;
  done;;

(* attente de synchronisation si besoin, 
   et initialisation de la nouvelle vague *)
let begin_sync n_s =
  Printf.printf "%ssynchronisation...\n" who;
  flush stdout;
  while waves.(local) >= (!old_wave + wave_depth) do 
    refresh_old_wave()
  done;
  sync_list := ( n_s :: !sync_list );
  sync_type.(0) <- n_s;
  Mutex.lock e;
  env := Array.append [| !env_buffer |] ( Array.sub !env 0 (wave_depth-1) );
  incr wave;
  Printf.printf "%sje surf sur la vague %d\n" who !wave ;
  flush stdout;
  Mutex.unlock e;;

(* decode un resultat *)

let decode r = match r with 
  Valeur v -> v
| Exn e -> raise (RemoteExn e);;
   
(* indique aux autres processus la fin de la vague *)
let end_sync () =
  Mutex.lock w;
  waves.(local) <- waves.(local) + 1;
  Mutex.unlock w;
  sync_type.(0) <- 0;
  ignore(Thread.create (refresh_waves) ())

(* attend que les autres processus aient fini avant de terminer *)
let close_Flight () =
  Printf.printf "%sj'ai fini, mais j'attend les autres...\n" who;
  flush stdout;
  while (!old_wave <> waves.(local)) do 
    Thread.delay 1.;
    refresh_old_wave();
  done;
  Printf.printf "%svoila!\n" who;
  flush stdout;;

(* *)
let my_input_line fd =
  let s = " " and r = ref ""
  in
    while (ThreadUnix.read fd s 0 1 > 0) && s.[0] <> '\n' do
      r := !r ^ s
    done;
    !r;;

(* demande au processus i la valeur du get n_get de la vague courante *)
let request i n_get =
  let (ic,oc) = 
    ThreadUnix.open_connection (Unix.ADDR_INET(addrs.(i),port+i-local))
  in
    Printf.printf "%sje demande le get %d de la vague %d au processus %d\n"
      who n_get !wave i;
    flush stdout;
    let dem = (Marshal.to_string (local,!wave,n_get) []) ^ "\n"
    in
      output_string oc dem;
      flush oc;
      let res = input_line ic
      in
        Printf.printf "%sje recupre la valeur demande au processus %d\n" 
	  who i;
	flush stdout;
	Unix.shutdown_connection ic;
        res;;

(* compose request, from_string et decode *)
let drequest i n_get = 
  decode ( Marshal.from_string (request i n_get) 0);;
 
(* function service gerant l'appel d'un get o l'arrive  une vague *)
let fun_service sd sa =
  let line = my_input_line sd in
    match (Marshal.from_string line 0) with
      | (p_or_n, 0, 0) -> 
	  Unix.close sd;
	  if p_or_n == nodes
	  then start := true
	  else
	    begin
	      Mutex.lock w;
	      waves.(p_or_n) <- waves.(p_or_n) + 1;
	      Mutex.unlock w
	    end
      | (p, n_wave, n_get) -> 
	  begin
	    Printf.printf "%sle processus %d me demande le get %d de la vague %d\n" who p n_get n_wave; 
	    flush stdout;
	    while (n_wave > !wave) do () done;
	    Mutex.lock e;
	    let s = (List.nth !env.(!wave-n_wave) (n_get-1)) in
	    let f = (Marshal.from_string s 0 : 'a fermeture)
	    in
	      Mutex.unlock e;
	      sync_type.(1) <- ( List.nth (List.rev !sync_list) (n_wave-1) );
	      let v = 
		try
		  Valeur (f());
		with
		  | e -> Exn e
	      in
		sync_type.(1) <- 0;
		let res = (Marshal.to_string v []) ^ "\n" 
		in
		  Printf.printf "%sj'envoi au processus %d la valeur du get %d de la vague %d\n" who p n_get n_wave;
		  flush stdout;
                  ignore(ThreadUnix.write sd res 0 (String.length res));
		  Unix.close sd
	  end;;

(* lane le serveur *)
let run ()=
  let sock = ThreadUnix.socket Unix.PF_INET Unix.SOCK_STREAM 0
  in
    Unix.bind sock (Unix.ADDR_INET(addrs.(local),port));
    Unix.listen sock nodes;
    while (true) do
      let (s_descr,s_addr) = ThreadUnix.accept sock
      in
  	ignore(Thread.create (fun_service s_descr) s_addr);
    done;;

(* lancement du serveur de gestion des appels en tache de fond *)
ignore(Thread.create (run) ());;

Printf.printf "%sserveur demarr\n" who;;
flush stdout;; 

while (not !start) do () done;;




